/* User, channel and server routines.

   OperStats - Copyright (c) 2000 Darcy Grexton
   Contact: cygnus@uniserve.ca, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/operstats.h"

Channel *chanlist[1024];
User *userlist[1024];

/* Add a new user into memory. */
static User *new_user (const char *nick)
{
    User *user, **list;

    user = scalloc (sizeof (User), 1);
    if (!nick)
	nick = "";
    strscpy (user->nick, nick, NICKLEN);
    list = &userlist[NHASH(user->nick)];
    user->next = *list;
    if (*list)
	(*list)->prev = user;
    *list = user;
    usercnt++;
    userstat++;
    if (userstat > maxusercnt)
    {
	maxusercnt = userstat;
	maxusertime = time (NULL);
    }
    return user;
}

/* Change a user's nick */
static void change_user_nick (User *user, const char *nick)
{
    User **list;

    if (user->prev)
	user->prev->next = user->next;
    else
	userlist[NHASH(user->nick)] = user->next;
    if (user->next)
	user->next->prev = user->prev;
    user->nick[1] = 0;
    strscpy (user->nick, nick, NICKLEN);
    list = &userlist[NHASH(user->nick)];
    user->next = *list;
    user->prev = NULL;
    if (*list)
	(*list)->prev = user;
    *list = user;
}

/* Handle a server NICK command.
       av[0] = nick
       If a new user:
           av[1] = hop count
           av[2] = signon time
           av[3] = username
           av[4] = hostname
           av[5] = user's server
           av[6] = 1 = service 0 = user
           av[7] = user's real name
       Else:
           av[1] = time of change
 */
void do_nick (const char *source, int ac, char **av)
{
    User *user;

    if (!*source)
    {
	MSG *msg;

	/* This is a new user; create a User structure for it. */
	if (runflags & RUN_DEBUG)
	    debug ("New user: %s", av[0]);

	/* Check if this user matches any AKills */
	check_akill (av[0], av[3], av[4], av[7]);

	/* Allocate User structure and fill it in. */
	user = new_user (av[0]);
	user->signon = atol (av[2]);
	user->user = sstrdup (av[3]);
	user->host = sstrdup (av[4]);
	user->server = sstrdup (av[5]);
	user->real = sstrdup (av[7]);

	/* Show them the LogonMSGs */
	for (msg = msglist; msg; msg = msg->next)
	    notice (s_GlobalNoticer, av[0], msg->text);
    }
    else        
    {
	/* An old user changing nicks. */
	user = finduser (source);

	if (!user)
	{
	    if (runflags & RUN_DEBUG)
		debug ("Got nickchange for nonexistant user %s->%s", source,
		    av[0]);
	    return;
	}

	if (runflags & RUN_DEBUG)
	    debug ("%s changed nick to %s", source, av[0]);

	/* Update their timestamp */
	user->signon = atol (av[1]);
	change_user_nick (user, av[0]);
    }
}

/* Find a user by nick.  Return NULL if user could not be found. */
User *finduser (const char *nick)
{
    User *user;

    user = userlist[NHASH(nick)];
    while (user && stricmp (user->nick, nick))
	user = user->next;
    return user;
}

/* Return the Channel structure corresponding to the named channel, or
   NULL if the channel was not found.
 */
Channel *findchan (const char *chan)
{
    Channel *c;

    c = chanlist[CHASH(chan)];
    while (c)
    {
	if (!stricmp (c->name, chan))
	    return c;
	c = c->next;
    }
    return NULL;
}

/* Remove and free a User structure. */
void delete_user (User * user)
{
    struct u_chanlist *c, *c2;

    usercnt--;
    userstat--;

    free (user->user);
    free (user->host);
    free (user->real);
    free (user->server);

    c = user->chans;

    while (c)
    {
	c2 = c->next;
	chan_deluser (user, c->chan);
	free (c);
	c = c2;  
    }

    if (user->prev)
	user->prev->next = user->next;
    else
	userlist[NHASH(user->nick)] = user->next;
    if (user->next)
	user->next->prev = user->prev;

    free (user);
}

/* Handle a JOIN command.
      av[0] = channels to join
 */
void do_join (const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c, *c2;
    long count, mem;  

    user = finduser (source);

    if (!user)
    {
	if (runflags & RUN_DEBUG)
	    debug ("Got JOIN to %s from nonexistant user %s", av[0], source);
	return;
    }

    t = av[0];

    while (*(s = t))
    {
	t = s + strcspn (s, ",");
	if (*t)
	    *t++ = 0;
        if (*s == '0')
        {
	    c = user->chans;
	    while (c)
	    {
		c2 = c->next;
		chan_deluser (user, c->chan);
		if (runflags & RUN_DEBUG)
		    debug ("%s parted all channels");
		free (c);
		c = c2;
	    }
	    user->chans = NULL;
	    continue;
	}
	if (runflags & RUN_DEBUG)
	    debug ("%s joined %s", source, s);
	chan_adduser (user, s);
	c = smalloc (sizeof (*c));
	c->next = user->chans; 
	c->prev = NULL;
	if (user->chans)
	    user->chans->prev = c;
	user->chans = c;
	c->chan = findchan (s);
    }
}

/* Handle a PART command */
void do_part (const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c;
 
    user = finduser (source);

    if (!user)  
	return;

    t = av[0];

    while (*(s = t))
    {
	t = s + strcspn (s, ",");
	if (*t) 
	    *t++ = 0;
	if (runflags & RUN_DEBUG)
	    debug ("%s parted %s", source, s);
	for (c = user->chans; c && stricmp (s, c->chan->name); c = c->next)
	    ;  
	if (c)
	{
	    chan_deluser (user, c->chan);
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		user->chans = c->next;
	    free (c);
	}
    }
}

/* Handle a KICK command */
void do_kick (const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c;   
        
    if (is_one_of_mine (av[1]))
	return;

    t = av[1];

    while (*(s = t))
    {
	t = s + strcspn (s, ",");
	if (*t)
	    *t++ = 0;
	user = finduser (s);
	if (!user)
	{
	    if (runflags & RUN_DEBUG)
		debug ("Got KICK for nonexistant user %s from %s", user->nick,
		    av[0]);
	    continue;
	}
	if (runflags & RUN_DEBUG)
	    debug ("%s was kicked from %s", s, av[0]);
	for (c = user->chans; c && stricmp (av[0], c->chan->name); c = c->next)
	    ;
	if (c) 
	{
	    chan_deluser (user, c->chan);
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		user->chans = c->next;
	    free (c);
	}
    }
    if (!source)
	return;
}

/* Handle a QUIT command */
void do_quit (const char *source, int ac, char **av)
{    
    User *user;

    user = finduser (source);

    if (!user)
    {
	if (runflags & RUN_DEBUG)
	    debug ("Got QUIT for nonexistant user %s", user->nick);
	return;
    }

    if (runflags & RUN_DEBUG)
	debug ("%s QUIT", source);

    delete_user (user);
}

/* Add a user to a channel, creating the channel as necessary. */
void chan_adduser (User * user, const char *chan)
{
    Channel *c = findchan (chan);
    Channel **list;
    int newchan = !c;
    struct c_userlist *u;

    if (newchan)
    {
	/* Allocate pre-cleared memory */
	c = scalloc (sizeof (Channel), 1);
	strscpy (c->name, chan, sizeof (c->name));
	list = &chanlist[CHASH(c->name)];
	c->next = *list;
	if (*list)
	    (*list)->prev = c;
	*list = c;
	c->creation_time = time (NULL);
	u = smalloc (sizeof (struct c_userlist));
	u->next = c->users;
	u->prev = NULL;
	if (c->users)
	    c->users->prev = u;
	c->users = u;
	u->user = user;
	++chancnt;
	++chanstat;
	if (chanstat > maxchancnt)
	{
	    maxchancnt = chanstat;
	    maxchantime = time (NULL);
	}
    }
}

/* Remove a user from a channel, deleting the channel as necessary. */
void chan_deluser (User * user, Channel * c)
{
    struct c_userlist *u;
    int i;

    for (u = c->users; u && u->user != user; u = u->next)
	;

    if (!u)
    {
	if (runflags & RUN_DEBUG)
	    debug ("Got chan_deluser for %s not in %s!", user->nick, c->name);
    }
    else
    {
	if (u->next)
	    u->next->prev = u->prev;
	if (u->prev)
	    u->prev->next = u->next;
	else
	    c->users = u->next;
	free (u);
    }

    for (u = c->chanops; u && u->user != user; u = u->next)
	;

    if (u)
    {
	if (u->next)
	    u->next->prev = u->prev;
	if (u->prev)
	    u->prev->next = u->next;
	else
	    c->chanops = u->next;
	free (u);
    }

    for (u = c->voices; u && u->user != user; u = u->next)
	;

    if (u)
    {
	if (u->next)
	    u->next->prev = u->prev;
	if (u->prev)
	    u->prev->next = u->next;
	else
	    c->voices = u->next;
	free (u);
    }

    if (!c->users)
    {
	--chancnt;
	--chanstat;
	if (c->topic)
	    free (c->topic);
	if (c->key)
	    free (c->key);
	for (i = 0; i < c->bancount; ++i)
	    if (c->bans[i])
		free (c->bans[i]);
	if (c->bansize)
	    free (c->bans); 
	if (c->next)
	    c->next->prev = c->prev;
	if (c->prev)
	    c->prev->next = c->next;
	else
	    chanlist[CHASH(c->name)] = c->next;
	free (c);
    }
}

/* Handle a KILL command.
      av[0] = nick being killed
      av[1] = reason
 */
void do_kill (const char *source, int ac, char **av)
{
    User *user;

    user = finduser (av[0]);

    if (!user)
    {
	if (runflags & RUN_DEBUG)
	    debug ("Got kill for nonexistant user %s", user->nick);
	return;
    }

    if (runflags & RUN_DEBUG)
	debug ("%s Killed", av[0]);
    delete_user (user);
}

/* Handle a user modechange */
void do_umode (const char *source, int ac, char **av)
{
    User *user;
    char *s;
    int add = 1;

    if (stricmp (source, av[0]))
	return;

    user = finduser (source);

    if (!user)
    {
	if (runflags & RUN_DEBUG)
	    debug ("MODE for nonexistant user %s: %s", source, av[1]);
	return;
    }

    if (runflags & RUN_DEBUG)
	debug ("Changing mode for %s to %s", source, av[1]);

    s = av[1]; 

    while (*s)
    {
	switch (*s++)
	{
	    case '+':
		add = 1;
		break;
	    case '-':
		add = 0;
		break;
	    case 'r':
		add ? (user->mode |= UMODE_r) : (user->mode &= ~UMODE_r);
                break;
	    case 'i':
		add ? (user->mode |= UMODE_i) : (user->mode &= ~UMODE_i);
		break;
	    case 'w':
		add ? (user->mode |= UMODE_w) : (user->mode &= ~UMODE_w);
		break;
	    case 'g':
		add ? (user->mode |= UMODE_g) : (user->mode &= ~UMODE_g);
		break;
	    case 's':
		add ? (user->mode |= UMODE_s) : (user->mode &= ~UMODE_s);
		break;
	    case 'o':
		add ? (user->mode |= UMODE_o) : (user->mode &= ~UMODE_o);
		break;
	}
    }
}

/* Handle a channel modechange */
void do_cmode (const char *source, int ac, char **av)
{
    Channel *chan;
    struct c_userlist *u;
    User *user;
    char *s, *nick;
    int add = 1;
    char *modestr = av[1];

    chan = findchan (av[0]);

    if (!chan)
    {
	if (runflags & RUN_DEBUG)
	    debug ("MODE for nonexistant channel %s: %s", av[0], av[1]);
	return;
    }

    s = modestr;
    ac -= 2;
    av += 2;

    while (*s)
    {
	switch (*s++)
	{
	    case '+':
		add = 1;
		break;
	    case '-':
		add = 0;
		break;
	    case 'i':
		add ? (chan->mode |= CMODE_i) : (chan->mode &= ~CMODE_i);
		break;
	    case 'm':
		add ? (chan->mode |= CMODE_m) : (chan->mode &= ~CMODE_m);
		break;
	    case 'n':
		add ? (chan->mode |= CMODE_n) : (chan->mode &= ~CMODE_n);
		break;
	    case 'p':
		add ? (chan->mode |= CMODE_p) : (chan->mode &= ~CMODE_p);
		break;
	    case 's':
		add ? (chan->mode |= CMODE_s) : (chan->mode &= ~CMODE_s);
		break;
	    case 't':
		add ? (chan->mode |= CMODE_t) : (chan->mode &= ~CMODE_t);
		break;
	    case 'c':
		add ? (chan->mode |= CMODE_c) : (chan->mode &= ~CMODE_c);
		break;
	    case 'R':
		add ? (chan->mode |= CMODE_R) : (chan->mode &= ~CMODE_R);
		break;
	    case 'r':
		add ? (chan->mode |= CMODE_r) : (chan->mode &= ~CMODE_r);
		break;
	    case 'k': 
		if (--ac < 0)
		{
		    if (runflags & RUN_DEBUG)
			debug ("MODE %s: Missing parameter for %ck",
				chan->name, add ? '+' : '-');
		    break;
		}
		if (chan->key)
		{
		    free (chan->key);
		    chan->key = NULL;
		}
		if (add)
		    chan->key = sstrdup(*av++);
		break;
	    case 'l':
		if (add)
		{
		    if (--ac < 0)
		    {
			if (runflags & RUN_DEBUG)
			    debug ("MODE %s: Missing parameter for +l",
				chan->name);
			break;
		    }
		    chan->limit = atoi (*av++); 
		}
		else
		    chan->limit = 0;        
		break;
	    case 'b':
		if (--ac < 0)
		{
		    if (runflags & RUN_DEBUG)
			debug ("MODE %s: Missing parameter for %cb",
			    chan->name, add ? '+' : '-');
		    break;
		}
		if (add)
		{
		    if (chan->bancount >= chan->bansize)
		    {
			chan->bansize += 8;
			chan->bans = srealloc(chan->bans, sizeof(char *) *
			    chan->bansize);
		    }
		    chan->bans[chan->bancount++] = sstrdup(*av++);
		}
		else
		{
		    char **s = chan->bans;
		    int i = 0;

		    while (i < chan->bancount && strcmp (*s, *av))
		    {
			i++;
			s++;
		    }
		    if (i < chan->bancount)
		    {
			chan->bancount--;
			if (i < chan->bancount)
			memmove (s, s+1, sizeof(char *) * (chan->bancount-i));
		    }
		    av++;
		}
		break;
	    case 'o':
		if (--ac < 0)
		{
		    if (runflags & RUN_DEBUG)
			debug ("MODE %s: Missing parameter for %co",
			    chan->name, add ? '+' : '-');
		    break;
		}
		nick = *av++;
		if (add)
		{
		    for (u = chan->chanops; u && stricmp (u->user->nick, nick);
			u = u->next)
			;

		    if (u)
			break;

		    user = finduser (nick);

		    if (!user)
		    {
			if (runflags & RUN_DEBUG)
			    debug ("MODE %s: +o for nonexistant user %s",
				chan->name, nick);
			break;
		    }
		    if (runflags & RUN_DEBUG)
			debug ("MODE %s: +o %s", chan->name, nick);
		    u = smalloc(sizeof(*u));
		    u->next = chan->chanops;
		    u->prev = NULL;
		    if (chan->chanops)
			chan->chanops->prev = u;
		    chan->chanops = u;
		    u->user = user;
		}
		else
		{
		    for (u = chan->chanops; u && stricmp (u->user->nick, nick);
			u = u->next)
			;
		    if (!u)
			break;
		    if (u->next)
			u->next->prev = u->prev;
		    if (u->prev)
			u->prev->next = u->next;
		    else
			chan->chanops = u->next;
		    free (u);
		}
		break;
	    case 'v':
		if (--ac < 0)
		{
		    if (runflags & RUN_DEBUG)
			debug ("MODE %s: Missing parameter for %cv",
			    chan->name, add ? '+' : '-');
		    break;
		}
		nick = *av++;
		if (add)
		{
		    for (u = chan->voices; u && stricmp (u->user->nick, nick);
			u = u->next)
			;

		    if (u)
			break;

		    user = finduser (nick);

		    if (!user)
		    {
			if (runflags & RUN_DEBUG)
			    debug ("MODE %s: +v for nonexistant user %s",
				chan->name, nick);
			break;
		    }
		    if (runflags & RUN_DEBUG)
			debug ("MODE %s: +v %s", chan->name, nick);
		    u = smalloc(sizeof(*u));
		    u->next = chan->voices;
		    u->prev = NULL;
		    if (chan->voices)
			chan->voices->prev = u;
		    chan->voices = u;
		    u->user = user;
		}
		else
		{
		    for (u = chan->voices; u && stricmp (u->user->nick, nick);
			u = u->next)
			;
		    if (!u)
			break;
		    if (u->next)
			u->next->prev = u->prev;
		    if (u->prev)
			u->prev->next = u->next;
		    else
			chan->voices = u->next;
		    free (u);
		}
		break;
	}
    }
}

/* Is the given nick voiced (channel mode +v) on the given channel? */
int is_voiced (const char *nick, const char *chan)
{
    Channel *c = findchan (chan);
    struct c_userlist *u;
 
    if (!c)
	return 0;
    for (u = c->voices; u; u = u->next)
	if (!stricmp (u->user->nick, nick))
	    return 1;
    return 0;
}

/* Is the given nick a channel operator on the given channel? */
int is_chanop (const char *nick, const char *chan)
{
    Channel *c = findchan (chan);
    struct c_userlist *u;

    if (!c)
	return 0;
    for (u = c->chanops; u; u = u->next)
	if (!stricmp (u->user->nick, nick))
	    return 1;
    return 0;
}

/* Return channel statistics - This is mainly so StatServ's counts
   match lusers, which doesn't include +s channels. This returns one
   of two things: The total amount of channels that aren't +s, or the
   total amount that are.
 */
int countchans (int secret)
{
    Channel *c;
    int i, total = 0;

    for (i = 0; i < 1024; i++)
    {
	for (c = chanlist[i]; c; c = c->next)
	{
	    if (secret && (c->mode & CMODE_s))
		++total;
	    else if (!secret && !(c->mode & CMODE_s))
		++total;
	}
    }
    return total;
}

/* Remove a user from the network. */
void kill_user (const char *source, const char *user, const char *reason)
{
    char *av[2];
    char buf[BUFSIZE];

    if (!user || !*user)
	return;

    if (!source || !*source)
	source = me.name;

    if (!reason)
	reason = "Killed";

    snprintf (buf, sizeof (buf), "%s (%s)", source, reason);

    av[0] = sstrdup (user);
    av[1] = buf;

    send_cmd (source, "KILL %s :%s", user, av[1]);
    do_kill (source, 2, av);
    free (av[0]);
}
